==> ado_wrapper-0.0.9-py3-none-any.whl: ado_wrapper/state_managed_abc.py <==
from typing import Any, TYPE_CHECKING, Literal, Callable
from dataclasses import dataclass, fields
from datetime import datetime

from ado_wrapper.utils import (
    get_resource_variables, extract_id, get_internal_field_names,
    ResourceAlreadyExists, DeletionFailed, ResourceNotFound, UpdateFailed,  # fmt: skip
)
from ado_wrapper.plan_resources.plan_resource import PlannedStateManagedResource

if TYPE_CHECKING:
    from ado_wrapper.client import AdoClient


def recursively_convert_to_json(attribute_name: str, attribute_value: Any) -> tuple[str, Any]:
    if isinstance(attribute_value, dict):
        return attribute_name, {key: recursively_convert_to_json("", value)[1] for key, value in attribute_value.items()}
    if isinstance(attribute_value, list):
        return attribute_name, [recursively_convert_to_json(attribute_name, value)[1] for value in attribute_value]
    if isinstance(attribute_value, datetime):
        return f"{attribute_name}::datetime", attribute_value.isoformat()
    if type(attribute_value) in get_resource_variables().values():
        class_name = str(type(attribute_value)).rsplit(".", maxsplit=1)[-1].removesuffix("'>")
        return attribute_name + "::" + class_name, attribute_value.to_json()
    return attribute_name, str(attribute_value)


def recursively_convert_from_json(dictionary: dict[str, Any]) -> Any:
    data_copy = dict(dictionary.items())  # Deep copy
    for key, value in dictionary.items():
        if isinstance(key, str) and "::" in key and key.split("::")[-1] != "datetime":
            instance_name, class_type = key.split("::")
            class_ = get_resource_variables()[class_type]
            del data_copy[key]
            data_copy[instance_name] = class_.from_json(value)
        elif isinstance(key, str) and key.endswith("::datetime"):
            del data_copy[key]
            data_copy[key.split("::")[0]] = datetime.fromisoformat(value)
    return data_copy


# ==========================================================================================


@dataclass
class StateManagedResource:
    @classmethod
    def from_request_payload(cls, data: dict[str, Any]) -> "StateManagedResource":
        raise NotImplementedError

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> "StateManagedResource":
        return cls(**recursively_convert_from_json(data))

    def to_json(self) -> dict[str, Any]:
        attribute_names = [field_obj.name for field_obj in fields(self)]
        attribute_values = [getattr(self, field_obj.name) for field_obj in fields(self)]
        combined = zip(attribute_names, attribute_values)
        return dict(recursively_convert_to_json(attribute_name, attribute_value) for attribute_name, attribute_value in combined)  ####

    @classmethod
    def get_by_id(cls, ado_client: "AdoClient", resource_id: str) -> "StateManagedResource":
        raise NotImplementedError

    @classmethod
    def get_by_url(cls, ado_client: "AdoClient", url: str) -> "StateManagedResource":
        if not url.startswith("https://"):
            url = f"https://dev.azure.com/{ado_client.ado_org}{url}"
        request = ado_client.session.get(url)
        if request.status_code == 404:
            raise ResourceNotFound(f"No {cls.__name__} found with that identifier!")
        if request.status_code >= 300:
            raise ValueError(f"Error getting {cls.__name__} by id: {request.text}")
        if "value" in request.json():
            return cls.from_request_payload(request.json()["value"][0])
        return cls.from_request_payload(request.json())

    @classmethod
    def create(
        cls, ado_client: "AdoClient", url: str, payload: dict[str, Any] | None = None
    ) -> "StateManagedResource | PlannedStateManagedResource":
        if ado_client.plan_mode:
            return PlannedStateManagedResource.create(cls, ado_client, url, payload)
        if not url.startswith("https://"):
            url = f"https://dev.azure.com/{ado_client.ado_org}" + url
        request = ado_client.session.post(url, json=payload or {})  # Create a brand new dict
        if request.status_code == 401:
            raise PermissionError(f"You do not have permission to create this {cls.__name__}!")
        if request.status_code == 409:
            raise ResourceAlreadyExists(f"The {cls.__name__} with that identifier already exist!")
        resource = cls.from_request_payload(request.json())
        ado_client.state_manager.add_resource_to_state(cls.__name__, extract_id(resource), resource.to_json())  # type: ignore[arg-type]
        return resource

    @classmethod
    def delete_by_id(cls, ado_client: "AdoClient", url: str, resource_id: str) -> None:
        """Deletes an object by its id. The id is passed so it can be removed from state"""
        if not url.startswith("https://"):
            url = f"https://dev.azure.com/{ado_client.ado_org}{url}"
        request = ado_client.session.delete(url)
        if request.status_code != 204:
            if request.status_code == 404:
                print("[ADO_WRAPPER] Resource not found, probably already deleted, removing from state")
            else:
                raise DeletionFailed(f"[ADO_WRAPPER] Error deleting {cls.__name__} ({resource_id}): {request.json()['message']}")
        ado_client.state_manager.remove_resource_from_state(cls.__name__, resource_id)  # type: ignore[arg-type]

    def delete(self, ado_client: "AdoClient") -> None:
        return self.delete_by_id(ado_client, extract_id(self))  # type: ignore[call-arg]  # pylint: disable=no-value-for-parameter

    def update(self, ado_client: "AdoClient", update_action: Literal["put", "patch"], url: str,  # pylint: disable=too-many-arguments
               attribute_name: str, attribute_value: Any, params: dict[str, Any]) -> None:  # fmt: skip
        """The params should be a dictionary which will be combined with the internal name and value of the attribute to be updated."""
        interal_names = get_internal_field_names(self.__class__)
        if attribute_name not in get_internal_field_names(self.__class__):
            raise ValueError(f"The attribute `{attribute_name}` is not editable!  Editable attributes are: {list(interal_names.keys())}")
        params |= {interal_names[attribute_name]: attribute_value}

        if ado_client.plan_mode:
            return PlannedStateManagedResource.update(self, ado_client, url, attribute_name, attribute_value, params)

        if not url.startswith("https://"):
            url = f"https://dev.azure.com/{ado_client.ado_org}{url}"
        request = ado_client.session.request(update_action, url, json=params)
        if request.status_code != 200:
            raise UpdateFailed(
                f"Failed to update {self.__class__.__name__} with id {extract_id(self)} and attribute {attribute_name} to {attribute_value}. \nReason:\n{request.text}"
            )
        setattr(self, attribute_name, attribute_value)
        ado_client.state_manager.update_resource_in_state(self.__class__.__name__, extract_id(self), self.to_json())  # type: ignore[arg-type]

    @classmethod
    def get_all(cls, ado_client: "AdoClient", url: str) -> list["StateManagedResource"]:
        if not url.startswith("https://"):
            url = f"https://dev.azure.com/{ado_client.ado_org}{url}"
        request = ado_client.session.get(url)
        if request.status_code >= 300:
            raise ValueError(f"Error getting all {cls.__name__}: {request.text}")
        return [cls.from_request_payload(resource) for resource in request.json()["value"]]

    @classmethod
    def get_by_abstract_filter(
        cls, ado_client: "AdoClient", func: Callable[["StateManagedResource"], bool]
    ) -> "StateManagedResource | None":
        """Used internally for getting resources by a filter function. The function should return True if the resource is the one you want."""
        resources = cls.get_all(ado_client)  # type: ignore[call-arg]  # pylint: disable=no-value-for-parameter
        for resource in resources:
            if func(resource):
                return resource
        return None

    def set_lifecycle_policy(self, ado_client: "AdoClient", policy: Literal["prevent_destroy", "ignore_changes"]) -> None:
        self.life_cycle_policy = policy  # TODO
        ado_client.state_manager.update_lifecycle_policy(self.__class__.__name__, extract_id(self), policy)  # type: ignore[arg-type]

==> ado_wrapper-1.10.0-py3-none-any.whl: ado_wrapper/state_managed_abc.py <==
from dataclasses import dataclass, fields
from datetime import datetime
from typing import TYPE_CHECKING, Any, Callable, Literal

from ado_wrapper.plan_resources.plan_resource import PlannedStateManagedResource
from ado_wrapper.errors import DeletionFailed, ResourceAlreadyExists, ResourceNotFound, UpdateFailed, InvalidPermissionsError  # fmt: skip
from ado_wrapper.utils import extract_id, get_internal_field_names, get_resource_variables

if TYPE_CHECKING:
    from ado_wrapper.client import AdoClient


def recursively_convert_to_json(attribute_name: str, attribute_value: Any) -> tuple[str, Any]:
    if isinstance(attribute_value, dict):
        return attribute_name, {key: recursively_convert_to_json("", value)[1] for key, value in attribute_value.items()}
    if isinstance(attribute_value, list):
        return attribute_name, [recursively_convert_to_json(attribute_name, value)[1] for value in attribute_value]
    if isinstance(attribute_value, datetime):
        return f"{attribute_name}::datetime", attribute_value.isoformat()
    if type(attribute_value) in get_resource_variables().values():
        class_name = str(type(attribute_value)).rsplit(".", maxsplit=1)[-1].removesuffix("'>")
        return attribute_name + "::" + class_name, attribute_value.to_json()
    return attribute_name, str(attribute_value)


def recursively_convert_from_json(dictionary: dict[str, Any]) -> Any:
    data_copy = dict(dictionary.items())  # Deep copy
    for key, value in dictionary.items():
        if isinstance(key, str) and "::" in key and key.split("::")[-1] != "datetime":
            instance_name, class_type = key.split("::")
            class_ = get_resource_variables()[class_type]
            del data_copy[key]
            data_copy[instance_name] = class_.from_json(value)
        elif isinstance(key, str) and key.endswith("::datetime"):
            del data_copy[key]
            data_copy[key.split("::")[0]] = datetime.fromisoformat(value)
    return data_copy


# ==========================================================================================


@dataclass
class StateManagedResource:
    @classmethod
    def from_request_payload(cls, data: dict[str, Any]) -> "StateManagedResource":
        raise NotImplementedError

    @classmethod
    def from_json(cls, data: dict[str, Any]) -> "StateManagedResource":
        return cls(**recursively_convert_from_json(data))

    def to_json(self) -> dict[str, Any]:
        attribute_names = [field_obj.name for field_obj in fields(self)]
        attribute_values = [getattr(self, field_obj.name) for field_obj in fields(self)]
        combined = zip(attribute_names, attribute_values)
        return dict(recursively_convert_to_json(attribute_name, attribute_value) for attribute_name, attribute_value in combined)

    @classmethod
    def get_by_id(cls, ado_client: "AdoClient", resource_id: str) -> "StateManagedResource":
        raise NotImplementedError

    @classmethod
    def get_by_url(cls, ado_client: "AdoClient", url: str) -> "StateManagedResource":
        if not url.startswith("https://"):
            url = f"https://dev.azure.com/{ado_client.ado_org}{url}"
        request = ado_client.session.get(url)
        if request.status_code == 404:
            raise ResourceNotFound(f"No {cls.__name__} found with that identifier!")
        if request.status_code >= 300:
            raise ValueError(f"Error getting {cls.__name__} by id: {request.text}")
        if "value" in request.json():
            return cls.from_request_payload(request.json()["value"][0])
        return cls.from_request_payload(request.json())

    @classmethod
    def create(
        cls, ado_client: "AdoClient", url: str, payload: dict[str, Any] | None = None, refetch: bool = False
    ) -> "StateManagedResource | PlannedStateManagedResource":
        """When creating, often the response doesn't contain all the data, refetching does a .get_by_id() after creation."""
        # If it already exists:
        # if cls.get_by_id(ado_client, extract_unique_name(payload)):
        #     raise ResourceAlreadyExists(f"The {cls.__name__} with that identifier already exist!")
        #     <update the resource>
        if ado_client.plan_mode:
            return PlannedStateManagedResource.create(cls, ado_client, url, payload)
        if not url.startswith("https://"):
            url = f"https://dev.azure.com/{ado_client.ado_org}" + url
        request = ado_client.session.post(url, json=payload or {})  # Create a brand new dict
        if request.status_code >= 300:
            if request.status_code in [401, 403]:
                raise InvalidPermissionsError(f"You do not have permission to create this {cls.__name__}! {request.text}")
            if request.status_code == 409:
                raise ResourceAlreadyExists(f"The {cls.__name__} with that identifier already exist!")
            raise ValueError(f"Error creating {cls.__name__}: {request.status_code} - {request.text}")
        resource = cls.from_request_payload(request.json())
        if refetch:
            resource = cls.get_by_id(ado_client, extract_id(resource))
        ado_client.state_manager.add_resource_to_state(cls.__name__, extract_id(resource), resource.to_json())  # type: ignore[arg-type]
        return resource

    @classmethod
    def delete_by_id(cls, ado_client: "AdoClient", url: str, resource_id: str) -> None:
        """Deletes an object by its id. The id is passed so it can be removed from state"""
        if not url.startswith("https://"):
            url = f"https://dev.azure.com/{ado_client.ado_org}{url}"
        request = ado_client.session.delete(url)
        if request.status_code != 204:
            if request.status_code == 404:
                if not ado_client.suppress_warnings:
                    print("[ADO_WRAPPER] Resource not found, probably already deleted, removing from state")
            else:
                if "message" in request.json():
                    raise DeletionFailed(f"[ADO_WRAPPER] Error deleting {cls.__name__} ({resource_id}): {request.json()['message']}")
                raise DeletionFailed(f"[ADO_WRAPPER] Error deleting {cls.__name__} ({resource_id}): {request.text}")
        ado_client.state_manager.remove_resource_from_state(cls.__name__, resource_id)  # type: ignore[arg-type]

    def delete(self, ado_client: "AdoClient") -> None:
        return self.delete_by_id(ado_client, extract_id(self))  # type: ignore[call-arg]  # pylint: disable=no-value-for-parameter

    def update(self, ado_client: "AdoClient", update_action: Literal["put", "patch"], url: str,  # pylint: disable=too-many-arguments
               attribute_name: str, attribute_value: Any, params: dict[str, Any]) -> None:  # fmt: skip
        """The params should be a dictionary which will be combined with the internal name and value of the attribute to be updated."""
        interal_names = get_internal_field_names(self.__class__)
        if attribute_name not in get_internal_field_names(self.__class__):
            raise ValueError(f"The attribute `{attribute_name}` is not editable!  Editable attributes are: {list(interal_names.keys())}")
        params |= {interal_names[attribute_name]: attribute_value}

        if ado_client.plan_mode:
            return PlannedStateManagedResource.update(self, ado_client, url, attribute_name, attribute_value, params)

        if not url.startswith("https://"):
            url = f"https://dev.azure.com/{ado_client.ado_org}{url}"
        request = ado_client.session.request(update_action, url, json=params)
        if request.status_code != 200:
            raise UpdateFailed(
                f"Failed to update {self.__class__.__name__} with id {extract_id(self)} and attribute {attribute_name} to {attribute_value}. \nReason:\n{request.text}"
            )
        setattr(self, attribute_name, attribute_value)
        ado_client.state_manager.update_resource_in_state(self.__class__.__name__, extract_id(self), self.to_json())  # type: ignore[arg-type]

    @classmethod
    def get_all(cls, ado_client: "AdoClient", url: str) -> list["StateManagedResource"]:
        if not url.startswith("https://"):
            url = f"https://dev.azure.com/{ado_client.ado_org}{url}"
        request = ado_client.session.get(url)
        if request.status_code >= 300:
            raise ValueError(f"Error getting all {cls.__name__}: {request.text}")
        return [cls.from_request_payload(resource) for resource in request.json()["value"]]

    @classmethod
    def get_by_abstract_filter(
        cls, ado_client: "AdoClient", func: Callable[["StateManagedResource"], bool]
    ) -> "StateManagedResource | None":
        """Used internally for getting resources by a filter function. The function should return True if the resource is the one you want."""
        resources = cls.get_all(ado_client)  # type: ignore[call-arg]  # pylint: disable=no-value-for-parameter
        for resource in resources:
            if func(resource):
                return resource
        return None

    def set_lifecycle_policy(self, ado_client: "AdoClient", policy: Literal["prevent_destroy", "ignore_changes"]) -> None:
        self.life_cycle_policy = policy  # TODO
        ado_client.state_manager.update_lifecycle_policy(self.__class__.__name__, extract_id(self), policy)  # type: ignore[arg-type]

    # def __enter__(self) -> "StateManagedResource":
    #     return self

    # def __exit__(self, *_: Any) -> None:
    #     self.delete(self.ado_client)
